C**********************************************************************
C                                                                     C
C     Program circle                                                  C
C                                                                     C
C     This is 'circle'; it reads particle positions and radii from    C
C     'position1.out' and outputs circumference coordinates to        C
C     'circles.dat' in a format suitable for gnuplot or unigraph.     C
C     It also calculates average coodination number and void space    C
C     and writes these to 'coord.out'                                 C
C                                                                     C
C     The resolution of the coordinates in controlled by the          C
C     parameter 'steps'                                               C
C                                                                     C
C     The size of the included area relative to that of the           C
C     simulation box is governed by parameter 'inset'                 C
C                                                                     C
C     Setting epsilon > 0 gives non-spherical particles               C
C                                                                     C
C**********************************************************************
C
       program circle
       implicit none
C
C---------------------------------------------------------------------
C  Initialise variables
C
       integer counter, i, xtract
       integer partmax, type, max
       integer inner, top(2), fract
       integer xtract, circl, coor
       integer grcal, nbr, vrn
       integer G05DYF
       parameter (max = 5000)
       character name*15
C
       double precision x(max), z(max), R(max), o(max)
       double precision xi(max), zi(max), Ri(max), oi(max)
       double precision radians, dummy
       double precision xmax, zmax, mean_u, mean_l
       double precision sigma, dR, dRmax
       double precision epsilon, inset
       double precision PI, area
       parameter (PI = 3.141592654, radians = PI/180.0D0)
C
       external G05CCF, G05DYF
C
       common /particles/ x(max), z(max), R(max), o(max)
       common /bounds/    mean_u, mean_l, xmax, zmax
       common /shapes/    epsilon, inset
       common /rad/       dR, dRmax
C
       call G05CCF(0)
C
C---------------------------------------------------------------------
C  Open files
C
       open(12,file='circles.dat')
       open(13,file='coord.out')
       open(15,file='cell.out')
       open(16,file='count.out')
C
C----------------------------------------------------------------------
C  Get options from file 'circle.opt'
C
       open(14,file='circle.opt')
C
       read(14,*) name
       read(14,*) xtract, circl, coor, fract
       read(14,*) grcal, nbr, vrn
C
       close(14)
C
C----------------------------------------------------------------------
C  Read data from 'name'
C
       open(11,file=name)
C
       read(11,*) counter
       do 3, i = 1, counter
        read(11,*) dummy, xi(i), zi(i), Ri(i), oi(i)
3      continue
C
       read(11,*) xmax, zmax, mean_u, mean_l
       read(11,*) sigma, dR, dRmax, inset
       read(11,*) partmax, type, epsilon
C
C----------------------------------------------------------------------
C  Write header info to 'coord.out'
C
       write(13,*) xmax, zmax, mean_u, mean_l
       write(13,*) sigma, dR, dRmax, inset
       write(13,*) partmax, type, epsilon
       write(13,*)
C
C---------------------------------------------------------------------
C  Isolate inner area of simulation box to eliminate edge effects
C
       if (xtract.eq.0)  then
          call edge(xi, zi, Ri, oi, counter, inner, top)
       else
          call extract(xi, zi, Ri, oi, counter, inner, top)
       end  if
C
C---------------------------------------------------------------------
C  Calculate coordinates of points on circumference;
C    output to 'circles.dat'
C
       if (circl.ne.0)  then
          call circ(inner)
       end  if
C
C---------------------------------------------------------------------
C  Calculate average coordination number.
C
       if (coor.ne.0)  then
          call coord(inner)
       end  if
C
C---------------------------------------------------------------------
C  Calculate radial distribution plot, G(r)
C
       if (grcal.ne.0)  then
          if (type.eq.1) dR = ((mean_u+mean_l)/2.0D0)*dR
          call gr_calc(inner)
       end  if
C
C---------------------------------------------------------------------
C  Calculate number of nearest neighbours
C
       if (vrn .ne. 0 .or. nbr .ne. 0)  then
          call voron(top, inner, vrn, nbr)
       end  if
C
C---------------------------------------------------------------------
C  Calculate packing fraction statistics
C
       if (fract.ne.0)  then
          call fraction(inner)
       end  if
C
C---------------------------------------------------------------------
C  Close files and finish
C
       write(6,*) 'Done'
       close(11)
       close(12)
       close(13)
       close(16)
C
       stop
       end
C*****************************************************************
C
C  Subroutine circ
C
C    Purpose :  Calculate points on circumference of circle
C 
C    Input Arguments : 
C
C         x, z   Coordinates of particle centres
C         R, o   Particle radii and orientations
C         inner  Number of particles considered
C
C    Output Arguments :
C
C         none     Writes to file 'circles.dat'
C
C    Common blocks :
C
C         /shapes/   epsilon, inset
C
C    Subroutines
C
C    Called from :
C
C          main
C
C    Calls to :
C
C          none
C
C*****************************************************************
       subroutine circ(inner)
       integer inner, degrees, steps, max
       parameter (max = 5000)
       parameter (degrees = 359, steps = 6)
       double precision x(max), z(max), R(max), o(max)
       double precision xdash, zdash, t, u
       double precision inset, epsilon
       double precision circle, angle, radius, R0
       double precision PI, radians
       parameter (PI = 3.141592654, radians = PI/180.0D0)
C
       common /particles/ x(max), z(max), R(max), o(max)
       common /shapes/ epsilon, inset
C
       write(6,*)
       write(6,*) 'Writing to circles.dat'
       write(6,*)
C
       do 15, i = 1, inner
        radius = (R(i)*epsilon)/2.0D0
        angle  = o(i) 
        do 17, j = 0, degrees, steps 
         circle = dble(j)*radians
         if (circle.gt.(degrees*radians))  then
            circle = circle - ((degrees+1)*radians)
         end  if
         R0 = radius+(radius*(sin(4.0D0*(circle+angle))))
         t = (R(i)-R0)*sin(circle)
         u = (R(i)-R0)*cos(circle) 
         xdash = t + x(i)
         zdash = u + z(i)
C         write(6,2001) xdash, zdash
         write(12,2001) xdash, zdash
17      continue
15     continue
C
       return
2001   format(f8.4, x, f8.4)
       end
C*****************************************************************
C
C  Subroutine link_len
C
C    Purpose :  Calculate distance between any two particle centres
C 
C    Input Arguments : 
C
C         x1, z1   Coordinated of particle centres
C         x2, z2
C
C    Output Arguments :
C
C         hypo     Distance between particle centres
C
C*****************************************************************
       subroutine link_len(x1, z1, x2, z2, hypo)
       double precision x1, z1, x2, z2, hypo
       double precision xdistance, zdistance
C
       xdistance = (x2-x1)**2
       zdistance = (z2-z1)**2
       hypo = sqrt(xdistance+zdistance)
C
       return
       end
C*****************************************************************
C
C  Subroutine coord
C
C    Purpose :  Calculate average coordination number
C 
C    Input Arguments :
C
C         x, z     Coordinates of particle centres
C         counter  Number of particles
C         inset    
C
C    Output Arguments :
C
C         none   (writes to file 'coord.out')
C
C*****************************************************************
       subroutine coord(count)
       integer count, i, j, max
       integer top, incr, num
       integer pos, cont, total
       parameter (max = 5000)
       integer contact(max), histo(0:10)
       double precision x(max), z(max), R(max), o(max)
       double precision Rerror, hypo
       double precision dtot, dcontact, dcount
       double precision xcent, zcent, Rcent
       double precision xcheck, zcheck, Rcheck
       double precision mean_u, mean_l, xmax, zmax
       double precision epsilon, inset, dover
       double precision xbot, zbot, xtop, ztop
       double precision x_out, x_in, z_out, z_in
       double precision stddev, val, var
       parameter (Rerror = 1.0D-2)
C
       common /particles/ x(max), z(max), R(max), o(max)
       common /bounds/    mean_u, mean_l, xmax, zmax
       common /shapes/    epsilon, inset
C
       write(6,*) 'Calculating average coordination number ...'
C
       xbot  = inset*2.0D0
       zbot  = xbot
       x_out = 0.0D0
       z_out = 0.0D0
       do 5, i = 1, count
        x_in = x(i)
        z_in = z(i)
        if (x_in.gt.x_out)  x_out = x_in
        if (z_in.gt.z_out)  z_out = z_in
5      continue        

       xtop = x_out-xbot
       ztop = z_out-zbot
C
       do 10, i = 1, count
        xcent = x(i)
        zcent = z(i)
        Rcent = R(i)
        if (xcent.gt.xbot.and.xcent.lt.xtop)  then
           if (zcent.gt.zbot.and.zcent.lt.ztop)  then
              incr = incr + 1
              do 8, j = 1, count
               if (i.ne.j)  then
                  xcheck = x(j)
                  zcheck = z(j)
                  Rcheck = Rcent+R(j)+Rerror
                  call link_len(xcent, zcent, xcheck, zcheck, hypo)
                  if (hypo.le.Rcheck.and.hypo.gt.0.0D0)  then
                     contact(incr) = contact(incr) + 1
                  end  if
               end  if
8             continue
           end  if
        end  if
10     continue
C
       do 13, i = 1, incr
C        write(6,*) i, contact(i)
        dcontact = dble(contact(i)) + dcontact
13     continue
C
       dcount   = dble(incr)
       dtot     = dcontact/dcount
C
C  Calculate std deviation and model coord number
C
       top = 0
       num = 0
       do 15, i = 1, incr
        val = (dble(contact(i))-dtot)**2
        var = var + val
        pos = contact(i)
        if (pos.gt.top)  top = pos
        histo(pos) = histo(pos) + 1
15     continue
       stddev = sqrt(var/incr)
C
       write(6,*)
       write(6,*) 'Coordination histogram :-'
       write(13,*)
       write(13,*) 'Coordination histogram :-'
       write(13,*)
C
       pos = 0
       do 17, i = 0, top
        pos = histo(i)
        total = total + pos
        write(6,*) i, pos
        write(13,2003) i, pos
        if (pos.gt.num)  then
           num  = pos
           cont = i
        end  if
17     continue
C
       dover = 100.0D0*(dble(num)/dble(total))
C
       write(13,*)
       write(13,*) '1. Ave coordination number'
       write(13,*) '2. Standard Deviation'
       write(13,*)
       write(13,2002) dtot, stddev
       write(6,*)
       write(6,*) 'Ave. coord no. and Std Dev. :-'
       write(6,2002) dtot, stddev
       write(6,2004) cont, num, dover
       write(6,*)
C
       return
2002   format(x, f4.2, 3x, f7.3)
2003   format(x, i2, 3x, i5)
2004   format('Modal coord no ',i2,' occurred ',i4,' times',
&                 ' (', f4.1, '%)')
       end
C*****************************************************************
C
C  Subroutine gr_calc
C
C    Purpose :  Calculate G(r) plot (as per Allen and Tildesley)
C               using dR radial increments.  Output results to 
C               file 'coord.out' (suitable for gnuplot).
C 
C    Input Arguments : 
C
C         area     relative area of particles in simulation box
C         counter  total number of particles
C         dRmax    max. radial distance to be evaluated
C         top      z position of 'highest' particle 
C
C    Output Arguments :
C
C         none
C
C    Common Blocks :
C
C         /bounds/    mean_u, mean_l, xmax, zmax
C         /rad/       dR, dRmax
C         /shapes/    epsilon, inset
C
C    Subroutines :
C
C    called from :
C
C         main
C
C    calls to :
C
C         link_len
C
C*****************************************************************
       subroutine gr_calc(counter)
       integer counter, times, max
       integer dRtop, incr, ihisto
       integer bin, histogram(10000)
       parameter (max = 5000)
C
       double precision x(max), z(max), R(max), o(max)
       double precision upper, lower, Gr, steps
       double precision hypo, xpos1, xpos2, zpos1, zpos2
       double precision rad, dhisto(max)
       double precision dR, dRmax, xmax, zmax, mean_u, mean_l
       double precision epsilon, inset, PI
       double precision xside, zside, factor
       double precision Fr(max), Ginf, Gxz
       double precision first, second, third, fourth, total
       double precision fifth, sixth, seventh, eighth
       parameter (PI = 3.141592654)
C
       common /particles/ x(max), z(max), R(max), o(max)
       common /bounds/    mean_u, mean_l, xmax, zmax
       common /rad/       dR, dRmax
       common /shapes/    epsilon, inset
C
C  Write screen headers
C
       write(6,*) 'Calculating Particle radial distribution function.'
       write(6,*)
C
C   Calculate G(r) as per Bell and Allen/Tildlesly
C
       rad = 0.0D0
       dRtop = int(dRmax/dR)
C       write(6,*) dRtop
       do 2000, times = 1, dRtop
C        write(6,*) times
        incr = incr + 1
        rad = dR*times
        do 1012, i = 1, (counter-1)
         xpos1 = x(i)
         zpos1 = z(i)
         do 1015, j = (i+1), counter
C          if (i.ne.j)  then
             xpos2 = x(j)
             zpos2 = z(j)
             call link_len(xpos1, zpos1, xpos2, zpos2, hypo)       
             if (hypo.le.rad)  then
                histogram(times) = histogram(times)+2
             end  if
C          end  if
1015     continue
1012    continue
2000   continue
C
C       write(6,*)
C
       do 10, i = 2, incr
        dhisto(i) = dble(histogram(i))/dble(counter)
        ihisto = histogram(i)-histogram(i-1)
C        write(6,*) i, ihisto
C        write(16,2002) ((i-1)*dR), ihisto
10     continue
C
C       write(6,*)
C
       xside = xmax - (inset*2.0D0)
       zside = zmax - (inset*2.0D0)
C
C  Apply scaling factor and write G(r) histogram to 'count.out'
C
       times = int(dRmax/dR)
       do 1017, bin = 1, times
        steps     = steps + dR
        Ginf      = 2.0D0*PI*steps
        factor    = sqrt(xside**2 + zside**2)
        if (steps.le.xside.and.steps.le.zside)  then
           first   = PI/2.0D0
           second  = steps/xside
           third   = steps/zside
           fourth  = (steps**2)/(2.0D0*xside*zside)
           total   = first-second-third+fourth
        else  if (steps.ge.xside.and.steps.le.zside)  then
           first   = PI/2.0D0
           second  = steps/xside
           third   = acos(xside/steps)
           fourth  = xside/(2.0D0*zside)
           fifth   = sqrt((steps**2/xside**2)-1)
           total   = first-second-third-fourth
        else  if (steps.ge.zside.and.steps.le.xside)  then
           first   = PI/2.0D0
           second  = steps/zside
           third   = acos(zside/steps)
           fourth  = zside/(2.0D0*xside)
           fifth   = sqrt((steps**2/zside**2)-1)
           total   = first-second-third-fourth
        else if (steps.ge.xside.and.steps.ge.zside.and.
&                steps.le.factor)  then
           first   = PI/20.D0
           second  = acos(zside/steps)
           third   = acos(xside/steps)
           fourth  = zside/(2.0D0*xside)
           fifth   = xside/(2.0D0*zside)
           sixth   = sqrt((steps**2/xside**2)-1)
           seventh = sqrt((steps**2/zside**2)-1)
           eighth  = steps**2/(2.0D0*xside*zside)
           total   = first-second-third-fourth-fifth-
&                    sixth-seventh-eighth
        else
           total   = 0.0D0
        end  if
        Gxz       = 4.0D0*steps*total
        Fr(bin)   = Ginf/Gxz
1017   continue
C
       steps = dR
       do 1020, bin = 2, times
        steps  = steps+dR
        upper  = dhisto(bin)/Fr(bin)
        lower  = dhisto(bin-1)/Fr(bin-1)
C        write(6,*) upper, lower, Fr(bin), Fr(bin-1)
        Gr     = (upper-lower)
        if (Gr.lt.0.0D0)  Gr = 0.0D0
C        write(6,2002)  steps, Gr 
        write(16,2002) steps, Gr
1020   continue
C
       write(6,*)
       write(6,*) '---------'
       return
2002   format(f4.1, 2x, f8.4)
       end
C*****************************************************************
C
C  Subroutine pack
C
C    Purpose :  
C
C    Input Arguments : 
C
C         x, z     coordinates of particle centres
C         R        particle radii
C         ave_dens areal fraction of particles in simulation box
C         count    total number of particles
C
C    Output Arguments :
C
C         none    writes to file 'area.out'
C
C    Common Blocks :
C
C         /bounds/    mean_u, mean_l, xmax, zmax
C         /shapes/    epsilon, inset
C
C    Subroutines :
C
C    called from :
C
C         main
C
C    calls to :
C
C         link_len
C
C*****************************************************************
       subroutine pack(count, ave_dens)
       integer count, times, particle
       integer rows, cols, dAtop
       integer marked, total, G05DYF
       integer hit, pos(100), max
       integer bin, histogram(100), bintot
       parameter (times = 100, max = 5000)
C
       double precision x(max), z(max), R(max), o(max)
       double precision poros(times), ave_dens
       double precision hist
       double precision xu_bound, zu_bound, size
       double precision x_u, z_u, resol
       double precision xpart, zpart, hypo
       double precision xpos1, zpos1, xpos2, zpos2
       double precision mean_u, mean_l, xmax, zmax
       double precision epsilon, inset, dist
       double precision fraction, dA, dAmax
       parameter (size = 5.0D0, resol = 1.0D-1)
       parameter (dA = 2.5D-2, dAmax = 1.0D0)
C
       common /particles/ x(max), z(max), R(max), o(max)
C
       external G05DYF
C
       common /bounds/ mean_u, mean_l, xmax, zmax
       common /shapes/ epsilon, inset
C
C  Write screen headers
C
       write(6,*) 'Packing density distribution function'
       write(6,*)
C
C   Calculate h(A) as per Dickinson et al, 
C    Ind Eng Chem Res 27(10) 1988 1941-46
C
       dAtop = int(dAmax/dA)
       xu_bound = xmax-inset
       zu_bound = zmax-inset
       do 100, i = 1, times
        write(6,*) i
        hit      = 0
        marked   = 0
        total    = 0
        particle = G05DYF(1, count)
        xpart    = x(particle)
        zpart    = z(particle)
        x_u      = xpart+size
        z_u      = zpart+size
        if (x_u.gt.xu_bound)  then
           x_u = xu_bound
           xpart = x_u - size
        end  if
        if (z_u.gt.zu_bound)  then
           z_u = zu_bound
           zpart = z_u - size
        end  if
        do 90, j = 1, count
         if (x(j).le.x_u.and.x(j).ge.xpart)  then
            if (z(j).le.z_u.and.z(j).ge.z_l)  then
               hit = hit + 1
               pos(hit) = j
             end  if
         end  if
90      continue
        cols = int(size/resol)
        rows = cols
        do 97, j = 1, rows
         do 95, k = 1, cols
          xpos1 = xpart+dble((k-1)*resol)
          zpos1 = zpart+dble((j-1)*resol)
C          write(6,*) k, j, xpos1, zpos1
          total = total + 1
          do 93, l = 1, hit
           xpos2 = x(pos(l))
           zpos2 = z(pos(l))
           dist  = R(pos(l))
           call link_len(xpos1, zpos1, xpos2, zpos2, hypo)
           if (hypo.lt.dist)  then
C              write(6,*) xpos1, zpos1, xpos2, zpos2
              marked = marked + 1
           end  if
93        continue
95       continue
97      continue
        fraction = dble(marked)/dble(total)
        poros(i) = (1.0D0-fraction)/ave_dens
        write(6,*) fraction, ave_dens
        bin = int(poros(i)/dA)+1
        if (bin.le.dAtop)  histogram(bin) = histogram(bin)+1
        bintot = bintot + 1
100    continue
C
C  Write h(A) histogram
C
       do 200, i = 1, dAtop
        hist = dble(histogram(i))/dble(bintot)
        write(15,2002) i*dA, hist
200    continue
C
       return
2002   format(f5.3, 2x, f3.2)
       end
C*****************************************************************
C
C  Subroutine edge
C
C    Purpose :  C
C 
C    Input Arguments : 
C
C         elap       number of iterations
C         counter    number of particle
C
C    Output Arguments
C
C         ideal_constant  for G(r) plot
C         area            for local packing density
C
C    Common Blocks :
C
C         /particles/ x_in, z_in, R_in, o_in
C         /bounds/    mean_u, mean_l, xmax, zmax
C         /shapes/    epsilon, inset
C         /moves/     xmove, zmove
C
C    Subroutines :
C
C    called from :
C
C         main
C
C    calls to :
C
C         link_len
C
C*****************************************************************
       subroutine edge(xi, zi, Ri, oi, counter, inner, top)
       integer max, counter, inner
       integer xtop, ztop, xbot, zbot, top(2)
       parameter (max = 5000)
C
       double precision x(max), z(max), R(max), o(max)
       double precision xi(max), zi(max), Ri(max), oi(max)
       double precision xmax, zmax, inset, epsilon
       double precision x_u, z_u, x_l, z_l, mean_u, mean_l
       double precision xcheck, zcheck, Rcheck, ocheck
       double precision xlow, zlow, xhigh, zhigh
       double precision xpos1, zpos1
       double precision xmove, zmove
C
       common /particles/ x(max), z(max), R(max), o(max)
       common /bounds/    mean_u, mean_l, xmax, zmax
       common /shapes/    epsilon, inset
       common /moves/     xmove, zmove
C
C---------------------------------------------------------------------
C  Extract particles within bounds
C
       x_u = xmax-inset
       z_u = zmax-inset
       x_l = inset
       z_l = inset
       do 15, i = 1, counter
        xcheck = xi(i)
        zcheck = zi(i)
        Rcheck = Ri(i)
        ocheck = oi(i)
        if (xcheck.ge.x_l.and.zcheck.ge.z_l)  then
           if (xcheck.le.x_u.and.zcheck.le.z_u)  then
              inner = inner+1
              x(inner) = xcheck
              z(inner) = zcheck
              R(inner) = Rcheck
              o(inner) = ocheck
           end  if
        end  if
15     continue
C
C---------------------------------------------------------------------
C  Reset origin and particle coordinates accordingly
C
       xhigh = 1000.0D0
       zhigh = 1000.0D0
       xlow  = 0.0D0
       zlow  = 0.0D0
       do 21, i = 1, inner
        xpos1 = x(i)
        zpos1 = z(i)
        if (xpos1.lt.xhigh)  then
           xbot = i
           xhigh = xpos1
        end  if
        if (zpos1.lt.zhigh)  then
           zbot = i
           zhigh = zpos1
        end  if
        if (xpos1.gt.xlow)  then
           xtop = i
           xlow = xpos1
        end  if
        if (zpos1.gt.zlow)  then
           ztop = i
           zlow = xpos1
        end  if
21     continue
C
       xmove = abs(x(xbot))+R(xbot)
       zmove = abs(z(zbot))+R(zbot)
C
       if (x(xbot).gt.0.0D0)  xmove = (2.0D0*R(xbot))-xmove
       if (z(zbot).gt.0.0D0)  zmove = (2.0D0*R(zbot))-zmove
C
       do 23, i = 1, inner
        x(i) = x(i)+xmove
        z(i) = z(i)+zmove
23     continue
C
       top(1) = xtop
       top(2) = ztop
C
       return
       end
C*****************************************************************
C
C  Subroutine extract
C
C    Purpose :  C
C 
C    Input Arguments : 
C
C         elap       number of iterations
C         counter    number of particle
C
C
C         ideal_constant  for G(r) plot
C         area            for local packing density
C
C    Common Blocks :
C
C         /particles/ x, z, R, o
C         /bounds/    mean_u, mean_l, xmax, zmax
C         /shapes/    epsilon, inset
C         /moves/     xmove, zmove
C
C    Subroutines :
C
C    called from :
C
C         main
C
C    calls to :
C
C         link_len
C
C*****************************************************************
       subroutine extract(xi, zi, Ri, oi, counter, inner, top)
       integer max, counter, inner
       integer xtop, ztop, xbot, zbot, top(2)
       parameter (max = 5000)
C
       double precision x(max), z(max), R(max), o(max)
       double precision xi(max), zi(max), Ri(max), oi(max)
       double precision xmax, zmax, inset, epsilon
       double precision x_u, z_u, x_l, z_l, mean_u, mean_l
       double precision xcheck, zcheck, Rcheck, ocheck
       double precision xhigh, zhigh, xlow, zlow
       double precision xpos1, zpos1
       double precision xmove, zmove, xbound, zbound
       double precision xstart, xstop, zstart, zstop, G05DAF
       parameter (xbound = 10.0D0, zbound = 10.0D0)
C
       common /particles/ x(max), z(max), R(max), o(max)
       common /bounds/    mean_u, mean_l, xmax, zmax
       common /shapes/    epsilon, inset
       common /moves/     xmove, zmove
C
C---------------------------------------------------------------------
C  Extract particles within bounds
C
       x_u = xmax-(xbound+inset)
       z_u = zmax-(zbound+inset)
       x_l = inset
       z_l = inset
       xstart = G05DAF(x_l, x_u)
       xstop  = xstart + xbound
       zstart = G05DAF(z_l, z_u)
       zstop  = zstart + zbound
C
       do 15, i = 1, counter
        xcheck = xi(i)
        zcheck = zi(i)
        Rcheck = Ri(i)
        ocheck = oi(i)
        if (xcheck.ge.xstart.and.zcheck.ge.zstart)  then
           if (xcheck.le.xstop.and.zcheck.le.zstop)  then
              inner = inner+1
              x(inner) = xcheck
              z(inner) = zcheck
              R(inner) = Rcheck
              o(inner) = ocheck
           end  if
        end  if
15     continue
C
C---------------------------------------------------------------------
C  Reset origin and particle coordinates accordingly
C
       xhigh = 1000.0D0
       zhigh = 1000.0D0
       xlow  = 0.0D0
       zlow  = 0.0D0
       do 21, i = 1, inner
        xpos1 = x(i)
        zpos1 = z(i)
        if (xpos1.lt.xhigh)  then
           xbot = i
           xhigh = xpos1
        end  if
        if (zpos1.lt.zhigh)  then
           zbot = i
           zhigh = zpos1
        end  if
        if (xpos1.gt.xlow)  then
           xtop = i
           xlow = xpos1
        end  if
        if (zpos1.gt.zlow)  then
           ztop = i
           zlow = zpos1
        end  if
21     continue
C
       xmove = abs(x(xbot))+R(xbot)
       zmove = abs(z(zbot))+R(zbot)
C
       if (x(xbot).gt.0.0D0)  xmove = (2.0D0*R(xbot))-xmove
       if (z(zbot).gt.0.0D0)  zmove = (2.0D0*R(zbot))-zmove
C
       do 23, i = 1, inner
        x(i) = x(i)+xmove
        z(i) = z(i)+zmove
23     continue
C
       top(1) = xtop
       top(2) = ztop
C
       return
       end
C*****************************************************************
C
C  Subroutine voron
C
C    Purpose :  
C 
C    Input Arguments : 
C
C         x1, z1   Coordinated of particle centres
C         x2, z2
C
C    Output Arguments :
C
C         hypo     Distance between particle centres
C
C    Subroutines :
C
C    Called from :
C
C         many
C
C    Calls to :
C
C         none
C
C*****************************************************************
       subroutine voron(top, inner, vrn, nbr)
       integer i, row, col, max, incr
       integer firstrow, lastrow, inner
       integer rowmax, colmax, maxcol, mincol
       integer nbr, num1, num2, diff, times
       integer resol, kilo, vrn
       parameter (max = 5000, kilo = 1000)
       parameter (resol = 5)
       integer grid(kilo, kilo), xcol(max), zrow(max)
       integer position(max), number(kilo)
       integer nbrs(kilo), particle
       integer count, id, histo(0:kilo)
       integer top(2), xtop, ztop
       integer first, last, hit
       integer nbrsmax, mode, modeval
C
       double precision drow, dcol, xpos1, zpos1
       double precision dtot, stddev, var, val
       double precision x(max), z(max), R(max), o(max)
C
       common /particles/ x(max), z(max), R(max), o(max)
C
C---------------------------------------------------------------------
C  Find box for each particle centre
C
       xtop = top(1)
       ztop = top(2)
C
       first = -1
       last  = 1
       incr  = 0
       drow = z(ztop)*dble(resol)
       dcol = x(xtop)*dble(resol)
       rowmax = int(drow)+1
       colmax = int(dcol)+1
C
C       write(6,*) rowmax, colmax, rowmax*colmax
C
       do 100, row = 1, rowmax
        do 99, col = 1, colmax
         xpos1 = dble(col)/dble(resol)
         zpos1 = dble(row)/dble(resol)
         call nearest(xpos1, zpos1, hit, inner)
         grid(col, row) = hit
C            write(6,*) row, col, grid(col, row)
99      continue
100    continue
C
C  Calculate voronoi polyhedra and output to cell.out
C
       if (vrn .eq. 1)  then
          call vorout(grid, rowmax, colmax)
          if (nbr .eq. 0)  return
       end  if
C
C If vrn is 0, we assume that nbr is 1, otherwise we wouldn't
C   be here.
C
       write(6,*) 'Calculating nearest neighbours ....'
       write(6,*)
C
       nbrsmax = 0
       do 1000, particle = 1, inner
        count    = 0
        maxcol   = 0
        mincol   = 1000
        incr     = 0
        do 10, row = 1, rowmax
         do 8, col = 1, colmax
          if (grid(col, row).eq.particle)  then
             incr = incr + 1
             xcol(incr) = col
             zrow(incr) = row
             if (xcol(incr).gt.maxcol)  maxcol = xcol(incr)
             if (xcol(incr).lt.mincol)  mincol = xcol(incr)
C             write(6,*) zrow(incr), xcol(incr)
          end  if
8        continue
10      continue
C        write(6,*) particle, incr
C
        if (incr.eq.0)  goto 1000
C
        firstrow = zrow(1)-1
        lastrow  = zrow(incr)+1
C
        if (firstrow.lt.1)      firstrow = 1
        if (lastrow.gt.rowmax)  lastrow = rowmax
C
C        write(6,*) firstrow, lastrow, mincol, maxcol
C
        do 15, row = firstrow, lastrow
         do 12, col = mincol, maxcol
          id = grid(col, row)
          if (id.ne.particle)  then
             count = count + 1
             position(count) = id
          end  if
12       continue
15      continue
C
C        write(6,*) particle, count
        times = 1
        call ihpsort(count, position)
        number(times) = position(1)
C
        do 18, i = 2, count
         num1 = position(i-1)
         num2 = position(i)
         diff = num2 - num1
         if (diff.ne.0) then
            times = times+1
            number(times) = num2
         end  if
18      continue
C
        do 20, i = 1, times
         if (number(i).ne.0)  nbrs(particle) = nbrs(particle) + 1
20      continue
C
C        write(6,*) particle, nbrs(particle)
C        write(6,*)
        tot = tot + nbrs(particle)
        histo(nbrs(particle)) = histo(nbrs(particle))+1
        if (nbrs(particle).gt.nbrsmax)  nbrsmax = nbrs(particle)
1000   continue
C
C  Calculate std deviation and number of nearest neighbours
C
       dtot = dble(tot)/dble(inner)
       do 1005, i = 1, inner
        val = (dble(nbrs(i))-dtot)**2
        var = var + val
1005   continue
       stddev = sqrt(var/dble(inner))
C
       write(6,*)
       write(6,*) 'Nearest Neighbours histogram :-'
       write(13,*)
       write(13,*) 'Nearest Neighbours histogram :-'
       write(13,*)
C
       mode    = 0
       modeval = 0
       do 1010, i = 1, nbrsmax
        if (histo(i).gt.modeval)  then
           mode    = i
           modeval = histo(i)
        end  if
        write(6,*) i, histo(i)
        write(13,2003) i, histo(i)
1010   continue
C
       write(13,*)
       write(13,*) '1. Ave nearest neighbours'
       write(13,*) '2. Std Deviation'
       write(13,*)
       write(13,2002) dtot, stddev
       write(6,*)
       write(6,*) 'Ave. nearest neighbours and Std Dev. :-'
       write(6,2002) dtot, stddev
       write(6,*) 'Modal NN ',mode,' occurred ',modeval,' times'
       write(6,*)
C
       return
2002   format(x, f4.2, 3x, f5.3)
2003   format(x, i2, 3x, i5)
       end
C*****************************************************************
C
C  Subroutine nearest
C
C    Purpose :  Calculates id of particle nearest to supplied point
C 
C    Input Arguments : 
C
C         x1, z1   Coordinates of point under consideration
C         inner    Number of particles in simulation
C
C    Output Arguments :
C
C         hit     number of nearest particle     
C
C    Subroutines :
C
C    Called from :
C
C         main
C
C    Calls to :
C
C         link_len
C
C    Common blocks:
C
C         /particles/  x, z, R, o
C
C*****************************************************************
       subroutine nearest(x1, z1, hit, inner)
       integer inner, max, hit
       parameter (max = 5000)
C
       double precision x1, z1
       double precision hypo, dist
       double precision xcent, zcent
       double precision x(max), z(max), R(max), o(max)
C
       common /particles/ x(max), z(max), R(max), o(max)
C
       dist = 1.0D6
       do 100, i = 1, inner
        xcent = x(i)
        zcent = z(i)
        call link_len(x1, z1, xcent, zcent, hypo)
        if (hypo.lt.dist)  then
           hit  = i
           dist = hypo
        end  if
100    continue
C
       return
       end
C*****************************************************************
C  Subroutine ihpsort
C
C    Purpose :  sorts particle list using HEAPSORT algorithm
C               (as per Press et al, Ch 8)
C               INTEGER VARIABLES ONLY
C 
C    Input Arguments : 
C
C         ra, rb, rc, rd  x, y, R and o values to be sorted
C                         ** ra is sort key **
C         n               no. of particles in sort list
C
C    Output Arguments
C
C         ra, rb, rc  sorted particle list
C
C    Subroutines :
C
C    Called from :
C
C         zposition
C
C    Calls to :
C
C         none
C
C*****************************************************************
       subroutine ihpsort(n, ra)
       integer n, ra(n)
       integer i, ir, j, l, rra
C
       if(n.lt.2)  return
       l = n/2+1
       ir = n
C
10     continue
        if (l.gt.1)  then
           l = l-1
           rra = ra(l)
        else
           rra = ra(ir)
           ra(ir) = ra(1)
           ir = ir-1
           if (ir.eq.1)  then
              ra(1) = rra
              return
           end  if
        end  if
        i = l
        j = l+l
20      if (j.le.ir)  then
           if (j.lt.ir)  then
              if (ra(j).lt.ra(j+1))  j = j+1
           end if
           if (rra.lt.ra(j))  then
              ra(i) = ra(j)
              i = j
              j = j+j
           else
              j = ir+1
           end  if
        goto 20
        end  if
        ra(i) = rra
       goto 10
       end
C*****************************************************************
C  Subroutine hpsort
C
C    Purpose :  sorts particle list using HEAPSORT algorithm
C               (as per Press et al, Ch 8)
C 
C    Input Arguments : 
C
C         ra   array of values to be sorted
C              ** ra is sort key **
C         n    no. of items in sort list
C
C    Output Arguments
C
C         ra   sorted array list
C
C    Subroutines :
C
C    Called from :
C
C         zposition
C
C    Calls to :
C
C         none
C
C*****************************************************************
       subroutine hpsort(n, ra)
       integer i, ir, j, l, n
       double precision ra(n), rra
C
       if(n.lt.2)  return
       l = n/2+1
       ir = n
C
10     continue
        if (l.gt.1)  then
           l = l-1
           rra = ra(l)
        else
           rra = ra(ir)
           ra(ir) = ra(1)
           ir = ir-1
           if (ir.eq.1)  then
              ra(1) = rra
              return
           end  if
        end  if
        i = l
        j = l+l
20      if (j.le.ir)  then
           if (j.lt.ir)  then
              if (ra(j).lt.ra(j+1))  j = j+1
           end if
           if (rra.lt.ra(j))  then
              ra(i) = ra(j)
              i = j
              j = j+j
           else
              j = ir+1
           end  if
        goto 20
        end  if
        ra(i) = rra
       goto 10
       end
C*****************************************************************
C
C  Subroutine vorout
C
C    Purpose :  Writes approx. voronoi diagram in gnuplot format
C               to file cell.out
C 
C    Input Arguments : 
C
C         grid      'Captured' array of particle id's
C         rowmax    No. of row elements in grid
C         colmax    No. of column elements in grid
C
C    Output Arguments :
C
C         none      writes to file 'cell.out'
C
C    Subroutines :
C
C    Called from :
C
C         nbours
C
C    Calls to :
C
C         none
C
C*****************************************************************
       subroutine vorout(grid, rowmax, colmax)
       integer grid(1000, 1000)
       integer overlay(1000, 1000)
       integer position(4)
       integer row, col
       integer row1, column1, row2, column2
       integer i, j, det
       integer empty, full, diff
       integer rowmax, colmax
       parameter (empty = 0, full = 1)
C
       write(6,*) 'Calculating voronoi polyhedra ...'
       write(6,*)
C
       do 100, row = 1, (rowmax-1)
        do 99, col = 1, (colmax-1)
         row1    = row
         row2    = row + 1
         column1 = col
         column2 = col + 1
         position(1) = grid(column1, row1)
         position(2) = grid(column2, row1)
         position(3) = grid(column2, row2)
         position(4) = grid(column1, row2)
C         write(6,*) position(1), position(2), position(3), position(4)
         do 95, i = 1, 3
          do 93, j = k+1, 4
           diff = position(j) - position(i)
C           write(6,*) diff
           if (diff.ne.0)  det = 1
93        continue
95       continue
         if (det.eq.0)  then
            overlay(column1, row1) = empty
         else
            overlay(column1, row1) = full
         end  if
C         write(16,*) row1, column1, overlay(column1, row1)
         det = 0
99      continue
100    continue
C
       do 115, i = 1, (rowmax-1)
        do 113, j = 1, (colmax-1)
         if (overlay(j, i).eq.full)  then
            write(15,*) j, i
         end  if
113     continue
115    continue
       return
       end
C*****************************************************************
C
C  Subroutine fraction
C
C    Purpose :  Calculates packing fraction and associated stats.
C 
C    Input Arguments : 
C
C         grid      'Captured' array of particle id's
C         rowmax    No. of row elements in grid
C         colmax    No. of column elements in grid
C
C    Output Arguments :
C
C         none      writes to file 'cell.out'
C
C    Common Blocks :
C
C         /particles/ x, z, R, o
C         /bounds/    mean_u, mean_l, xmax, zmax
C         /shapes/    epsilon, inset
C         /moves/     xmove, zmove   
C
C    Subroutines :
C
C    Called from :
C
C         main
C
C    Calls to :
C
C         none
C
C*****************************************************************
       subroutine fraction(inner)
       integer G05DYF
       integer MC_runs, partmax, type
       integer max, inner, areabox
       integer i, j, k, xtract
       integer median, modul
       integer hit, steps, increments
       integer row, col, rowmax, colmax
       integer number, times, particle
       integer modemax, mode
       integer ival, valmax, histo(1000)
       parameter (max = 5000, MC_runs = 100)
       parameter (increments = 10, xtract = 0)
C
       double precision x(max), z(max), R(max), o(max)
       double precision area(1000), area_allup(10)
       double precision areamedian, box
       double precision areamax, areamin, areamean
       double precision maxtot, mintot, meantot
       double precision medtot, devtot, dval
       double precision x_u, x_l, z_u, z_l
       double precision x_uR, z_uR
       double precision xpos1, zpos1
       double precision xpos2, zpos2, Rpos2
       double precision xmax, zmax, mean_u, mean_l, resol
       double precision sigma, dR, dRmax, epsilon, inset
       double precision hypo, xrand, zrand, PI
       double precision dhit, dtimes, xmove, zmove
       double precision val, var, stddev, areamode
       double precision modetot
       parameter (PI = 3.141592654, box = 2.5D0)
       parameter (resol = 1.0D-1)
C
       character*14 name(50)
C
       common /particles/ x(max), z(max), R(max), o(max)
       common /bounds/    mean_u, mean_l, xmax, zmax
       common /shapes/    epsilon, inset
       common /moves/     xmove, zmove
C
       open(11,file='box.name')
       open(14,file='areabox.out')
C
       write(6,*) 'Calculating packing fraction ...'
C
C---------------------------------------------------------------------
C  Write output file header.
C
       write(14,*) ' Mean    Min    Max  Median  Std Dev'
       write(14,*)
C
C-----------------------------------------------------------------------
C   Read data from 'box.name'
C
       read(11,*) number
       do 3, i = 1, number
        read(11,*) name(i)
3      continue
       do 5, times = 1, number
C
C-----------------------------------------------------------------------
C   Clear stats variables.
C
        areamin    = 0.0D0
        areamax    = 0.0D0
        areamedian = 0.0D0
        areamean   = 0.0D0
        val        = 0.0D0
        var        = 0.0D0
        stddev     = 0.0D0
        hit        = 0
        do 6, i = 1, 100
         area(i) = 0.0D0
6       continue
C
C-----------------------------------------------------------------------
C  Calculate bounds
C
        x_uR = xmax - xmove
        z_uR = zmax - zmove
C
C=======================================================================
C-----------------------------------------------------------------------
C   Acquire particles centres within chosen area from particle list
C
        do 20, j = 1, MC_runs
         steps = mod(j, increments)
         if (steps.eq.0.or.j.eq.1)  then
            write(6,*) 'Done ',j, ' of ', MC_runs
         end  if
C
C-----------------------------------------------------------------------
C   Monte Carlo routine for calculating relative area of particles
C
         particle = G05DYF(1, inner)
C         write(6,*) inner, particle
         xrand    = x(particle)
         zrand    = z(particle)
         x_l = xrand-box
         x_u = xrand+box
         if (x_l.lt.0.0D0)  x_l = 0.0D0
         if (x_u.gt.x_uR)   x_u = x_uR
         z_l = zrand-box
         z_u = zrand+box
         if (z_l.lt.0.0D0)  z_l = 0.0D0
         if (z_u.gt.z_uR)   z_u = z_uR
C
         colmax  = int(box*2.0D0/resol)
         rowmax  = int(box*2.0D0/resol)
         areabox = rowmax*colmax
C
         do 14, row = 1, rowmax
          do 12, col = 1, colmax
           xpos1 = x_l+(col*resol)
           zpos1 = z_l+(row*resol)
           do 111, k = 1, inner
            xpos2 = x(k)
            zpos2 = z(k)
            Rpos2 = R(k)
            call link_len(xpos1, zpos1, xpos2, zpos2, hypo)
            if (hypo.le.Rpos2)  then
               hit = hit + 1
               goto 112
            end  if
111        continue
112        continue
12        continue
14       continue
C
C-----------------------------------------------------------------------
C   Calculate packing fraction.
C
         dhit    = dble(hit)
         dtimes  = dble(areabox)
         area(j) = dhit/dtimes
C         write(6,2002) area(j), (area(j)*100.0D0)
         hit = 0
20      continue
C=======================================================================
C-----------------------------------------------------------------------
C
C  Construct area histogram.
C
       do 21, i = 1, 100
        histo(i) = 0
21     continue
       valmax = 0
       do 22, i = 1, MC_runs
        dval = (area(i)*1000.0D0)/10.0D0
        ival = int(dval)+1
        if (ival.gt.valmax)  valmax = ival
        histo(ival) = histo(ival) + 1
22     continue
C
C  Find modal value
C
       modemax = 0
       do 23, i = 1, valmax
        if (histo(i).gt.modemax)  then
           modemax = histo(i)
           mode    = i
        end  if
23     continue
       areamode = dble(mode)/100.0D0
C
C  Calculate min, max, mean and median value of relative area, also
C    standard deviation of packing fraction.
C
        call hpsort(MC_runs, area)
        areamin  = area(1)
        areamax  = area(MC_runs)
C
        do 25, i = 1, MC_runs
         areamean = areamean+area(i)
25      continue
        areamean = areamean/dble(MC_runs)
C
        do 30, i = 1, MC_runs
         val = (areamean-area(i))**2
         var = var + val
30      continue
        stddev = sqrt(var/MC_runs)
C
        modul  = MC_runs/2
        median = mod(MC_runs, 2)
        if (median.ne.0) then
           areamedian = area(modul+1)
        else
           areamedian = (area(modul)+area(modul+1))/2
        end  if
C
        meantot = meantot + areamean
        maxtot  = maxtot  + areamax
        mintot  = mintot  + areamin
        medtot  = medtot  + areamedian
        modetot = modetot  + areamode
        devtot  = devtot  + stddev
C
        area_allup(times) = areamean
C
        write(14,2006) areamean, areamin, areamax, areamedian, stddev
        write(6,2004)  areamean, stddev
        write(6,2003)  areamin, areamax
        write(6,2005)  areamode, areamedian
        write(6,*)
C
        var = 0.0D0
C
        close(10)
5      continue
C
C-----------------------------------------------------------------------
C  Calculate all-up averages and standard deviation.
C
       meantot = meantot/number
       maxtot  = maxtot/number
       mintot  = mintot/number
       medtot  = medtot/number
       modetot = modetot/number
       devtot  = devtot/number
C
       var = 0.0D0
C
       do 50, i = 1, number
        val = (meantot-area_allup(i))**2
        var = var + val
50     continue
       stddev = sqrt(var/number)
C
C       write(6,*)
C       do 52, i = 1, valmax
C        write(6,*) i, histo(i)
C52     continue

       write(14,*)
       write(14,2006) meantot, mintot, maxtot, medtot, devtot
       write(14,2008) stddev
C
       write(14,*)
       write(14,*) 'Bed constants :-'
       write(14,*)
       write(14,*) xmax, zmax, mean_u, mean_l
       write(14,*) sigma, dR, dRmax, inset
       write(14,*) partmax, type, epsilon
C
       write(6,*) 'Overall values :-'
       write(6,2004) meantot, stddev
       write(6,2003) mintot, maxtot
       write(6,2005) modetot, medtot
       write(6,*)
       write(6,*) 'Done'
       write(6,*)
C
C-----------------------------------------------------------------------
C   Close files, return
C
       close(11)
       close(14)
2001   format (i3, 2x, f4.3)
2002   format ('Relative cell packing fraction is '
&              , f4.3, ' or ', f5.2, '%')
2003   format ('Min pf   : ', f4.3, x, 'Max pf    : ', f4.3)
2004   format ('Mean pf  : ', f4.3, x, 'Std Dev   : ', f4.3)
2005   format ('Modal pf : ', f4.3, x, 'Median pf : ', f4.3)
2006   format (x, f4.3, 3x, f4.3, 3x, f4.3, 3x, f4.3, 3x, f4.3)
2008   format (x, f4.3)
       return
       end
C*****************************************************************
